<?php

namespace App\Models;

use Psr\Http\Message\ServerRequestInterface as Request;

class Place extends BaseModel {

	/**
     * The table associated with the model.
     *
     * @var string
     */
   	protected $table = 'place';

   	/**
     * The attributes that are mass assignable.
     *
     * @var array
     */
   	protected $fillable = array(
        'name',
        'description',
        'address_1',
        'address_2',
        'phone',
        'website_url',
        'timetable',
        'disabled',
   	);

   	/**
     * The attributes that should be hidden for arrays.
     *
     * @var array
     */
    protected $hidden = array(
    );

    /**
     * Get the field created at.
     *
     * @return integer
     */
    public function getCreatedAtAttribute()
    {
        return $this->dateToTimestamp($this->attributes['created_at']);
    }

    /**
     * Set the field created at.
     *
     * @param  string  $value
     */
    public function setCreatedAtAttribute($value)
    {
        $this->attributes['created_at'] = $this->timestampToDate($value);
    }

    /**
     * Get the field updated at.
     *
     * @return integer
     */
    public function getUpdatedAtAttribute()
    {
        return $this->dateToTimestamp($this->attributes['updated_at']);
    }

    /**
     * Set the field updated at.
     *
     * @param  string  $value
     */
    public function setUpdatedAtAttribute($value)
    {
        $this->attributes['updated_at'] = $this->timestampToDate($value);
    }

    /**
     * Get the field disabled.
     *
     * @return boolean
     */
    public function getDisabledAttribute()
    {
        return $this->intToBoolean($this->attributes['disabled']);
    }

    /**
     * Get search cursor
     * @param  array $filter
     * @param  Request $request
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public static function getCursor($filter, Request $request) {

        unset($filter['_page']);
        unset($filter['_limit']);
        unset($filter['_order']);
        unset($filter['_sort']);

        $query = Place::query();

        // Filter name
        if (isset($filter['name'])) {
            $query = $query->where('name', 'LIKE', "%{$filter['name']}%");
        }

        // Filter by related categories
        if (isset($filter['categories'])) {
            $query = $query->whereHas('_categories', function(\Illuminate\Database\Eloquent\Builder $query) use ($filter) {
                $query->whereIn('place_category', is_array($filter['categories']) ? $filter['categories'] : [$filter['categories']], 'and');
            });
        }

        // Filter latitude
        if (isset($filter['latitude'])) {
            $query = $query->where('latitude', $filter['latitude']);
        }
        // Set min for latitude if defined
        if (isset($filter['latitude__min'])) {
            $query = $query->where('latitude', '>=', $filter['latitude__min']);
        }
        // Set max for latitude if defined
        if (isset($filter['latitude__max'])) {
            $query = $query->where('latitude', '<=', $filter['latitude__max']);
        }

        // Filter longitude
        if (isset($filter['longitude'])) {
            $query = $query->where('longitude', $filter['longitude']);
        }
        // Set min for longitude if defined
        if (isset($filter['longitude__min'])) {
            $query = $query->where('longitude', '>=', $filter['longitude__min']);
        }
        // Set max for longitude if defined
        if (isset($filter['longitude__max'])) {
            $query = $query->where('longitude', '<=', $filter['longitude__max']);
        }

        // Filter by related services
        if (isset($filter['services'])) {
            $query = $query->whereHas('_services', function(\Illuminate\Database\Eloquent\Builder $query) use ($filter) {
                $query->whereIn('service', is_array($filter['services']) ? $filter['services'] : [$filter['services']], 'and');
            });
        }

        // Filter by related owner
        if (isset($filter['owner'])) {
            $query = $query->where('owner', $filter['owner']);
        }

        // Filter disabled
        if (isset($filter['disabled'])) {
            $query = $query->where('disabled', static::toBoolean($filter['disabled']));
        }

        return $query;
    }

    /**
     * Return relationship for categories
     * @return \Illuminate\Database\Eloquent\Relations\BelongsToMany
     */
    public function _categories()
    {
        return $this->belongsToMany('App\Models\PlaceCategory', 'place__categories', 'place', 'place_category');
    }

    /**
     * Return relationship for services
     * @return \Illuminate\Database\Eloquent\Relations\BelongsToMany
     */
    public function _services()
    {
        return $this->belongsToMany('App\Models\Service', 'place__services', 'place', 'service');
    }

    /**
     * Return relationship for owner
     * @return \Illuminate\Database\Eloquent\Relations\HasOne
     */
    public function _owner()
    {
        return $this->hasOne('App\Models\User', '_id', 'owner');
    }
}